"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import pytest

import salt.modules.disk as disk
from tests.support.mock import MagicMock, patch


@pytest.fixture
def stub_disk_usage():
    return {
        "/": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
        "/dev": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
        "/run": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
        "/run/lock": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
        "/run/shm": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
        "/run/user": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
        "/sys/fs/cgroup": {
            "filesystem": None,
            "1K-blocks": 10000,
            "used": 10000,
            "available": 10000,
            "capacity": 10000,
        },
    }


@pytest.fixture
def stub_disk_inodeusage():
    return {
        "/": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
        "/dev": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
        "/run": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
        "/run/lock": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
        "/run/shm": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
        "/run/user": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
        "/sys/fs/cgroup": {
            "inodes": 10000,
            "used": 10000,
            "free": 10000,
            "use": 10000,
            "filesystem": None,
        },
    }


@pytest.fixture
def stub_disk_percent():
    return {
        "/": 50,
        "/dev": 10,
        "/run": 10,
        "/run/lock": 10,
        "/run/shm": 10,
        "/run/user": 10,
        "/sys/fs/cgroup": 10,
    }


@pytest.fixture
def stub_disk_blkid():
    return {"/dev/sda": {"TYPE": "ext4", "UUID": None}}


@pytest.fixture
def configure_loader_modules():
    return {disk: {}}


def test_usage_dict(stub_disk_usage):
    with patch.dict(disk.__grains__, {"kernel": "Linux"}), patch(
        "salt.modules.disk.usage", MagicMock(return_value=stub_disk_usage)
    ):
        mock_cmd = MagicMock(return_value=1)
        with patch.dict(disk.__salt__, {"cmd.run": mock_cmd}):
            assert stub_disk_usage == disk.usage(args=None)


def test_usage_none():
    with patch.dict(disk.__grains__, {"kernel": "Linux"}), patch(
        "salt.modules.disk.usage", MagicMock(return_value="")
    ):
        mock_cmd = MagicMock(return_value=1)
        with patch.dict(disk.__salt__, {"cmd.run": mock_cmd}):
            assert "" == disk.usage(args=None)


def test_inodeusage(stub_disk_inodeusage):
    with patch.dict(disk.__grains__, {"kernel": "OpenBSD"}), patch(
        "salt.modules.disk.inodeusage", MagicMock(return_value=stub_disk_inodeusage)
    ):
        mock = MagicMock()
        with patch.dict(disk.__salt__, {"cmd.run": mock}):
            assert stub_disk_inodeusage == disk.inodeusage(args=None)


def test_percent(stub_disk_percent):
    with patch.dict(disk.__grains__, {"kernel": "Linux"}), patch(
        "salt.modules.disk.percent", MagicMock(return_value=stub_disk_percent)
    ):
        mock = MagicMock()
        with patch.dict(disk.__salt__, {"cmd.run": mock}):
            assert stub_disk_percent == disk.percent(args=None)


def test_percent_args():
    with patch.dict(disk.__grains__, {"kernel": "Linux"}), patch(
        "salt.modules.disk.percent", MagicMock(return_value="/")
    ):
        mock = MagicMock()
        with patch.dict(disk.__salt__, {"cmd.run": mock}):
            assert "/" == disk.percent("/")


def test_blkid(stub_disk_blkid):
    with patch.dict(
        disk.__salt__, {"cmd.run_stdout": MagicMock(return_value=1)}
    ), patch("salt.modules.disk.blkid", MagicMock(return_value=stub_disk_blkid)):
        assert stub_disk_blkid == disk.blkid()


@pytest.mark.skip_on_windows(reason="Skip on Windows")
@pytest.mark.skip_on_darwin(reason="Skip on Darwin")
@pytest.mark.skip_on_freebsd
def test_blkid_token():
    run_stdout_mock = MagicMock(return_value={"retcode": 1})
    with patch.dict(disk.__salt__, {"cmd.run_all": run_stdout_mock}):
        disk.blkid(token="TYPE=ext4")
        run_stdout_mock.assert_called_with(
            ["blkid", "-t", "TYPE=ext4"], python_shell=False
        )


def test_dump():
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(disk.__salt__, {"cmd.run_all": mock}):
        disk.dump("/dev/sda")
        mock.assert_called_once_with(
            "blockdev --getro --getsz --getss --getpbsz --getiomin "
            "--getioopt --getalignoff --getmaxsect --getsize "
            "--getsize64 --getra --getfra /dev/sda",
            python_shell=False,
        )


def test_wipe():
    mock = MagicMock(return_value={"retcode": 0, "stdout": ""})
    with patch.dict(disk.__salt__, {"cmd.run_all": mock}):
        disk.wipe("/dev/sda")
        mock.assert_called_once_with("wipefs -a /dev/sda", python_shell=False)


def test_tune():
    mock = MagicMock(
        return_value=(
            "712971264\n512\n512\n512\n0\n0\n88\n712971264\n365041287168\n512\n512"
        )
    )
    with patch.dict(disk.__salt__, {"cmd.run": mock}):
        mock_dump = MagicMock(return_value={"retcode": 0, "stdout": ""})
        with patch("salt.modules.disk.dump", mock_dump):
            kwargs = {"read-ahead": 512, "filesystem-read-ahead": 1024}
            disk.tune("/dev/sda", **kwargs)

            mock.assert_called_with(
                "blockdev --setra 512 --setfra 1024 /dev/sda", python_shell=False
            )


def test_format():
    """
    unit tests for disk.format
    """
    device = "/dev/sdX1"
    mock = MagicMock(return_value=0)
    with patch.dict(disk.__salt__, {"cmd.retcode": mock}), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        assert disk.format_(device) is True


def test_fat_format():
    """
    unit tests for disk.format when using fat argument
    """
    device = "/dev/sdX1"
    expected = ["mkfs", "-t", "fat", "-F", 12, "/dev/sdX1"]
    mock = MagicMock(return_value=0)
    with patch.dict(disk.__salt__, {"cmd.retcode": mock}), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        assert disk.format_(device, fs_type="fat", fat=12) is True
        args, kwargs = mock.call_args_list[0]
        assert expected == args[0]


@pytest.mark.skip_if_binaries_missing("lsblk", "df", check_all=True)
def test_fstype():
    """
    unit tests for disk.fstype
    """
    device = "/dev/sdX1"
    fs_type = "ext4"
    mock = MagicMock(return_value="FSTYPE\n{}".format(fs_type))
    with patch.dict(disk.__grains__, {"kernel": "Linux"}), patch.dict(
        disk.__salt__, {"cmd.run": mock}
    ), patch("salt.utils.path.which", MagicMock(return_value=True)):
        assert disk.fstype(device) == fs_type


def test_resize2fs():
    """
    unit tests for disk.resize2fs
    """
    device = "/dev/sdX1"
    mock = MagicMock()
    with patch.dict(disk.__salt__, {"cmd.run_all": mock}), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        disk.resize2fs(device)
        mock.assert_called_once_with("resize2fs {}".format(device), python_shell=False)


@pytest.mark.skip_on_windows(reason="Skip on Windows")
@pytest.mark.skip_if_binaries_missing("mkfs")
def test_format_():
    """
    unit tests for disk.format_
    """
    device = "/dev/sdX1"
    mock = MagicMock(return_value=0)
    with patch.dict(disk.__salt__, {"cmd.retcode": mock}):
        disk.format_(device=device)
        mock.assert_any_call(["mkfs", "-t", "ext4", device], ignore_retcode=True)


@pytest.mark.skip_on_windows(reason="Skip on Windows")
@pytest.mark.skip_if_binaries_missing("mkfs")
def test_format__fat():
    """
    unit tests for disk.format_ with FAT parameter
    """
    device = "/dev/sdX1"
    mock = MagicMock(return_value=0)
    with patch.dict(disk.__salt__, {"cmd.retcode": mock}):
        disk.format_(device=device, fs_type="fat", fat=12)
        mock.assert_any_call(
            ["mkfs", "-t", "fat", "-F", 12, device], ignore_retcode=True
        )



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
import pytest

import salt.modules.win_timezone as win_timezone
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skipif(not win_timezone.HAS_PYTZ, reason="This test requires pytz"),
]


@pytest.fixture
def configure_loader_modules():
    return {
        win_timezone: {
            "__opts__": {},
            "__salt__": {},
            "__utils__": {},
        },
    }


def test_get_zone_normal():
    """
    Test if it get current timezone (i.e. Asia/Calcutta)
    """
    mock_read_ok = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "India Standard Time",
        }
    )
    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read_ok}):
        assert win_timezone.get_zone() == "Asia/Calcutta"


def test_get_zone_normal_dstoff():
    """
    Test if it gets current timezone with dst off (i.e. America/Denver)
    """
    mock_read_ok = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "Mountain Standard Time_dstoff",
        }
    )
    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read_ok}):
        assert win_timezone.get_zone() == "America/Denver"


def test_get_zone_normal_dstoff_issue():
    """
    Test regression with dstoff fix stripping unwanted characters
    """
    mock_read_ok = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "FLE Standard Time",
        }
    )
    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read_ok}):
        assert win_timezone.get_zone() == "Europe/Kiev"


@pytest.mark.parametrize("timezone", win_timezone.mapper.list_win())
def test_get_zone_all(timezone):
    """
    Test all Win zones are properly resolved and none returns Unknown
    """
    mock_read_ok = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": timezone,
        }
    )
    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read_ok}):
        assert win_timezone.get_zone() != "Unknown"


def test_get_zone_unknown():
    """
    Test get_zone with unknown timezone (i.e. Indian Standard Time)
    """
    mock_read_error = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "Indian Standard Time",
        }
    )
    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read_error}):
        assert win_timezone.get_zone() == "Unknown"


def test_get_zone_error():
    """
    Test get_zone when it encounters an error
    """
    mock_read_fatal = MagicMock(
        return_value={"pid": 78, "retcode": 1, "stderr": "", "stdout": ""}
    )
    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read_fatal}):
        with pytest.raises(CommandExecutionError):
            win_timezone.get_zone()


def test_get_offset():
    """
    Test if it get current numeric timezone offset from UCT (i.e. +0530)
    """
    mock_read = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "India Standard Time",
        }
    )

    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read}):
        assert win_timezone.get_offset() == "+0530"


def test_get_zonecode():
    """
    Test if it get current timezone (i.e. PST, MDT, etc)
    """
    mock_read = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "India Standard Time",
        }
    )

    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read}):
        assert win_timezone.get_zonecode() == "IST"


def test_set_zone():
    """
    Test if it unlinks, then symlinks /etc/localtime to the set timezone.
    """
    mock_write = MagicMock(
        return_value={"pid": 78, "retcode": 0, "stderr": "", "stdout": ""}
    )
    mock_read = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "India Standard Time",
        }
    )

    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_write}), patch.dict(
        win_timezone.__salt__, {"cmd.run_all": mock_read}
    ):
        assert win_timezone.set_zone("Asia/Calcutta")


def test_zone_compare():
    """
    Test if it checks the md5sum between the given timezone, and
    the one set in /etc/localtime. Returns True if they match,
    and False if not. Mostly useful for running state checks.
    """
    mock_read = MagicMock(
        return_value={
            "pid": 78,
            "retcode": 0,
            "stderr": "",
            "stdout": "India Standard Time",
        }
    )

    with patch.dict(win_timezone.__salt__, {"cmd.run_all": mock_read}):
        assert win_timezone.zone_compare("Asia/Calcutta")


def test_get_hwclock():
    """
    Test if it get current hardware clock setting (UTC or localtime)
    """
    assert win_timezone.get_hwclock() == "localtime"


def test_set_hwclock():
    """
    Test if it sets the hardware clock to be either UTC or localtime
    """
    assert not win_timezone.set_hwclock("UTC")



"""
Simple Smoke Tests for Connected Proxy Minion
"""
import logging
import random

import pytest

import salt.proxy.dummy
import salt.utils.path
import salt.utils.platform

log = logging.getLogger(__name__)


@pytest.fixture
def salt_call_cli(salt_proxy):
    return salt_proxy.salt_call_cli(timeout=120)


@pytest.mark.slow_test
def test_can_it_ping(salt_call_cli):
    """
    Ensure the proxy can ping
    """
    ret = salt_call_cli.run("test.ping")
    assert ret.returncode == 0, ret
    assert ret.data is True


@pytest.mark.slow_test
def test_list_pkgs(salt_call_cli):
    """
    Package test 1, really just tests that the virtual function capability
    is working OK.
    """
    ret = salt_call_cli.run("pkg.list_pkgs")
    assert ret.returncode == 0, ret
    for package_name in salt.proxy.dummy._initial_state()["packages"]:
        assert package_name in ret.data


@pytest.mark.slow_test
def test_upgrade(salt_call_cli):
    ret = salt_call_cli.run("pkg.upgrade")
    assert ret.returncode == 0, ret
    # Assert that something got upgraded
    assert ret.data
    assert ret.data["coreutils"]["new"] == "2.0"
    assert ret.data["redbull"]["new"], "1000.99"


@pytest.fixture
def service_name():
    return random.choice(list(salt.proxy.dummy._initial_state()["services"]))


@pytest.mark.slow_test
def test_service_list(salt_call_cli, service_name):
    ret = salt_call_cli.run("service.list")
    assert ret.returncode == 0, ret
    assert service_name in ret.data


@pytest.mark.slow_test
def test_service_start(salt_call_cli):
    ret = salt_call_cli.run("service.start", "samba")
    assert ret.returncode == 0, ret
    ret = salt_call_cli.run("service.status", "samba")
    assert ret.returncode == 0, ret
    assert ret.data is True


@pytest.mark.slow_test
def test_service_get_all(salt_call_cli, service_name):
    ret = salt_call_cli.run("service.get_all")
    assert ret.returncode == 0, ret
    assert service_name in ret.data


@pytest.mark.slow_test
def test_grains_items(salt_call_cli):
    ret = salt_call_cli.run("grains.items")
    assert ret.returncode == 0, ret
    assert ret.data["kernel"] == "proxy"
    assert ret.data["kernelrelease"] == "proxy"



import pytest
from pytestshellutils.utils.processes import terminate_process

import salt.ext.tornado.ioloop
import salt.utils.event
import salt.utils.stringutils


@pytest.mark.slow_test
def test_event_return(master_opts):
    evt = None
    try:
        evt = salt.utils.event.EventReturn(master_opts)
        evt.start()
    except TypeError as exc:
        if "object" in str(exc):
            pytest.fail("'{}' TypeError should have not been raised".format(exc))
    finally:
        if evt is not None:
            terminate_process(evt.pid, kill_children=True)



"""
Setup of Python virtualenv sandboxes.

.. versionadded:: 0.17.0
"""


import logging
import os

import salt.utils.functools
import salt.utils.platform
import salt.utils.versions
import salt.version
from salt.exceptions import CommandExecutionError, CommandNotFoundError

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "virtualenv"


def __virtual__():
    if "virtualenv.create" in __salt__:
        return __virtualname__
    return (False, "virtualenv module could not be loaded")


def managed(
    name,
    venv_bin=None,
    requirements=None,
    system_site_packages=False,
    distribute=False,
    use_wheel=False,
    clear=False,
    python=None,
    extra_search_dir=None,
    never_download=None,
    prompt=None,
    user=None,
    cwd=None,
    index_url=None,
    extra_index_url=None,
    pre_releases=False,
    no_deps=False,
    pip_download=None,
    pip_download_cache=None,
    pip_exists_action=None,
    pip_ignore_installed=False,
    proxy=None,
    use_vt=False,
    env_vars=None,
    no_use_wheel=False,
    pip_upgrade=False,
    pip_pkgs=None,
    pip_no_cache_dir=False,
    pip_cache_dir=None,
    process_dependency_links=False,
    no_binary=None,
    **kwargs
):
    """
    Create a virtualenv and optionally manage it with pip

    name
        Path to the virtualenv.

    venv_bin: virtualenv
        The name (and optionally path) of the virtualenv command. This can also
        be set globally in the minion config file as ``virtualenv.venv_bin``.

    requirements: None
        Path to a pip requirements file. If the path begins with ``salt://``
        the file will be transferred from the master file server.

    use_wheel: False
        Prefer wheel archives (requires pip >= 1.4).

    python: None
        Python executable used to build the virtualenv

    user: None
        The user under which to run virtualenv and pip.

    cwd: None
        Path to the working directory where `pip install` is executed.

    no_deps: False
        Pass `--no-deps` to `pip install`.

    pip_exists_action: None
        Default action of pip when a path already exists: (s)witch, (i)gnore,
        (w)ipe, (b)ackup.

    proxy: None
        Proxy address which is passed to `pip install`.

    env_vars: None
        Set environment variables that some builds will depend on. For example,
        a Python C-module may have a Makefile that needs INCLUDE_PATH set to
        pick up a header file while compiling.

    no_use_wheel: False
        Force to not use wheel archives (requires pip>=1.4)

    no_binary
        Force to not use binary packages (requires pip >= 7.0.0)
        Accepts either :all: to disable all binary packages, :none: to empty the set,
        or a list of one or more packages

    pip_upgrade: False
        Pass `--upgrade` to `pip install`.

    pip_pkgs: None
        As an alternative to `requirements`, pass a list of pip packages that
        should be installed.

    process_dependency_links: False
        Run pip install with the --process_dependency_links flag.

        .. versionadded:: 2017.7.0

    Also accepts any kwargs that the virtualenv module will. However, some
    kwargs, such as the ``pip`` option, require ``- distribute: True``.

    .. code-block:: yaml

        /var/www/myvirtualenv.com:
          virtualenv.managed:
            - system_site_packages: False
            - requirements: salt://REQUIREMENTS.txt
            - env_vars:
                PATH_VAR: '/usr/local/bin/'
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    if "virtualenv.create" not in __salt__:
        ret["result"] = False
        ret["comment"] = "Virtualenv was not detected on this system"
        return ret

    if salt.utils.platform.is_windows():
        venv_py = os.path.join(name, "Scripts", "python.exe")
    else:
        venv_py = os.path.join(name, "bin", "python")
    venv_exists = os.path.exists(venv_py)

    # Bail out early if the specified requirements file can't be found
    if requirements and requirements.startswith("salt://"):
        cached_requirements = __salt__["cp.is_cached"](requirements, __env__)
        if not cached_requirements:
            # It's not cached, let's cache it.
            cached_requirements = __salt__["cp.cache_file"](requirements, __env__)
        # Check if the master version has changed.
        if cached_requirements and __salt__["cp.hash_file"](
            requirements, __env__
        ) != __salt__["cp.hash_file"](cached_requirements, __env__):
            cached_requirements = __salt__["cp.cache_file"](requirements, __env__)
        if not cached_requirements:
            ret.update(
                {
                    "result": False,
                    "comment": "pip requirements file '{}' not found".format(
                        requirements
                    ),
                }
            )
            return ret
        requirements = cached_requirements

    # If it already exists, grab the version for posterity
    if venv_exists and clear:
        ret["changes"]["cleared_packages"] = __salt__["pip.freeze"](bin_env=name)
        ret["changes"]["old"] = __salt__["cmd.run_stderr"](
            "{} -V".format(venv_py)
        ).strip("\n")

    # Create (or clear) the virtualenv
    if __opts__["test"]:
        if venv_exists and clear:
            ret["result"] = None
            ret["comment"] = "Virtualenv {} is set to be cleared".format(name)
            return ret
        if venv_exists and not clear:
            ret["comment"] = "Virtualenv {} is already created".format(name)
            return ret
        ret["result"] = None
        ret["comment"] = "Virtualenv {} is set to be created".format(name)
        return ret

    if not venv_exists or (venv_exists and clear):
        try:
            venv_ret = __salt__["virtualenv.create"](
                name,
                venv_bin=venv_bin,
                system_site_packages=system_site_packages,
                distribute=distribute,
                clear=clear,
                python=python,
                extra_search_dir=extra_search_dir,
                never_download=never_download,
                prompt=prompt,
                user=user,
                use_vt=use_vt,
                **kwargs
            )
        except CommandNotFoundError as err:
            ret["result"] = False
            ret["comment"] = "Failed to create virtualenv: {}".format(err)
            return ret

        if venv_ret["retcode"] != 0:
            ret["result"] = False
            ret["comment"] = venv_ret["stdout"] + venv_ret["stderr"]
            return ret

        ret["result"] = True
        ret["changes"]["new"] = __salt__["cmd.run_stderr"](
            "{} -V".format(venv_py)
        ).strip("\n")

        if clear:
            ret["comment"] = "Cleared existing virtualenv"
        else:
            ret["comment"] = "Created new virtualenv"

    elif venv_exists:
        ret["comment"] = "virtualenv exists"

    # Check that the pip binary supports the 'use_wheel' option
    if use_wheel:
        min_version = "1.4"
        max_version = "9.0.3"
        cur_version = __salt__["pip.version"](bin_env=name)
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        too_high = salt.utils.versions.compare(
            ver1=cur_version, oper=">", ver2=max_version
        )
        if too_low or too_high:
            ret["result"] = False
            ret["comment"] = (
                "The 'use_wheel' option is only supported in "
                "pip between {} and {}. The version of pip detected "
                "was {}.".format(min_version, max_version, cur_version)
            )
            return ret

    # Check that the pip binary supports the 'no_use_wheel' option
    if no_use_wheel:
        min_version = "1.4"
        max_version = "9.0.3"
        cur_version = __salt__["pip.version"](bin_env=name)
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        too_high = salt.utils.versions.compare(
            ver1=cur_version, oper=">", ver2=max_version
        )
        if too_low or too_high:
            ret["result"] = False
            ret["comment"] = (
                "The 'no_use_wheel' option is only supported in "
                "pip between {} and {}. The version of pip detected "
                "was {}.".format(min_version, max_version, cur_version)
            )
            return ret

    # Check that the pip binary supports the 'no_binary' option
    if no_binary:
        min_version = "7.0.0"
        cur_version = __salt__["pip.version"](bin_env=name)
        too_low = salt.utils.versions.compare(
            ver1=cur_version, oper="<", ver2=min_version
        )
        if too_low:
            ret["result"] = False
            ret["comment"] = (
                "The 'no_binary' option is only supported in "
                "pip {} and newer. The version of pip detected "
                "was {}.".format(min_version, cur_version)
            )
            return ret

    # Populate the venv via a requirements file
    if requirements or pip_pkgs:
        try:
            before = set(__salt__["pip.freeze"](bin_env=name, user=user, use_vt=use_vt))
        except CommandExecutionError as exc:
            ret["result"] = False
            ret["comment"] = exc.strerror
            return ret

        if requirements:

            if isinstance(requirements, str):
                req_canary = requirements.split(",")[0]
            elif isinstance(requirements, list):
                req_canary = requirements[0]
            else:
                raise TypeError("pip requirements must be either a string or a list")

            if req_canary != os.path.abspath(req_canary):
                cwd = os.path.dirname(os.path.abspath(req_canary))

        pip_ret = __salt__["pip.install"](
            pkgs=pip_pkgs,
            requirements=requirements,
            process_dependency_links=process_dependency_links,
            bin_env=name,
            use_wheel=use_wheel,
            no_use_wheel=no_use_wheel,
            no_binary=no_binary,
            user=user,
            cwd=cwd,
            index_url=index_url,
            extra_index_url=extra_index_url,
            download=pip_download,
            download_cache=pip_download_cache,
            pre_releases=pre_releases,
            exists_action=pip_exists_action,
            ignore_installed=pip_ignore_installed,
            upgrade=pip_upgrade,
            no_deps=no_deps,
            proxy=proxy,
            use_vt=use_vt,
            env_vars=env_vars,
            no_cache_dir=pip_no_cache_dir,
            cache_dir=pip_cache_dir,
            **kwargs
        )
        ret["result"] &= pip_ret["retcode"] == 0
        if pip_ret["retcode"] > 0:
            ret["comment"] = "{}\n{}\n{}".format(
                ret["comment"], pip_ret["stdout"], pip_ret["stderr"]
            )

        after = set(__salt__["pip.freeze"](bin_env=name))

        new = list(after - before)
        old = list(before - after)

        if new or old:
            ret["changes"]["packages"] = {
                "new": new if new else "",
                "old": old if old else "",
            }
    return ret


manage = salt.utils.functools.alias_function(managed, "manage")



"""
Functions for working with files
"""


import codecs
import contextlib
import errno
import io
import logging
import os
import re
import shutil
import stat
import subprocess
import tempfile
import time
import urllib.parse

import salt.modules.selinux
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, FileLockError, MinionError
from salt.utils.decorators.jinja import jinja_filter

try:
    import fcntl

    HAS_FCNTL = True
except ImportError:
    # fcntl is not available on windows
    HAS_FCNTL = False

log = logging.getLogger(__name__)

LOCAL_PROTOS = ("", "file")
REMOTE_PROTOS = ("http", "https", "ftp", "swift", "s3")
VALID_PROTOS = ("salt", "file") + REMOTE_PROTOS
TEMPFILE_PREFIX = "__salt.tmp."

HASHES = {
    "sha512": 128,
    "sha384": 96,
    "sha256": 64,
    "sha224": 56,
    "sha1": 40,
    "md5": 32,
}
HASHES_REVMAP = {y: x for x, y in HASHES.items()}


def __clean_tmp(tmp):
    """
    Remove temporary files
    """
    try:
        rm_rf(tmp)
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            "Exception while removing temp directory: %s",
            exc,
            exc_info_on_loglevel=logging.DEBUG,
        )


def guess_archive_type(name):
    """
    Guess an archive type (tar, zip, or rar) by its file extension
    """
    name = name.lower()
    for ending in (
        "tar",
        "tar.gz",
        "tgz",
        "tar.bz2",
        "tbz2",
        "tbz",
        "tar.xz",
        "txz",
        "tar.lzma",
        "tlz",
    ):
        if name.endswith("." + ending):
            return "tar"
    for ending in ("zip", "rar"):
        if name.endswith("." + ending):
            return ending
    return None


def mkstemp(*args, **kwargs):
    """
    Helper function which does exactly what ``tempfile.mkstemp()`` does but
    accepts another argument, ``close_fd``, which, by default, is true and closes
    the fd before returning the file path. Something commonly done throughout
    Salt's code.
    """
    if "prefix" not in kwargs:
        kwargs["prefix"] = "__salt.tmp."
    close_fd = kwargs.pop("close_fd", True)
    fd_, f_path = tempfile.mkstemp(*args, **kwargs)
    if close_fd is False:
        return fd_, f_path
    os.close(fd_)
    del fd_
    return f_path


def recursive_copy(source, dest):
    """
    Recursively copy the source directory to the destination,
    leaving files with the source does not explicitly overwrite.

    (identical to cp -r on a unix machine)
    """
    for root, _, files in salt.utils.path.os_walk(source):
        path_from_source = root.replace(source, "").lstrip(os.sep)
        target_directory = os.path.join(dest, path_from_source)
        if not os.path.exists(target_directory):
            os.makedirs(target_directory)
        for name in files:
            file_path_from_source = os.path.join(source, path_from_source, name)
            target_path = os.path.join(target_directory, name)
            shutil.copyfile(file_path_from_source, target_path)


def copyfile(source, dest, backup_mode="", cachedir=""):
    """
    Copy files from a source to a destination in an atomic way, and if
    specified cache the file.
    """
    if not os.path.isfile(source):
        raise OSError("[Errno 2] No such file or directory: {}".format(source))
    if not os.path.isdir(os.path.dirname(dest)):
        raise OSError("[Errno 2] No such file or directory: {}".format(dest))
    bname = os.path.basename(dest)
    dname = os.path.dirname(os.path.abspath(dest))
    tgt = mkstemp(prefix=bname, dir=dname)
    shutil.copyfile(source, tgt)
    bkroot = ""
    if cachedir:
        bkroot = os.path.join(cachedir, "file_backup")
    if backup_mode == "minion" or backup_mode == "both" and bkroot:
        if os.path.exists(dest):
            backup_minion(dest, bkroot)
    if backup_mode == "master" or backup_mode == "both" and bkroot:
        # TODO, backup to master
        pass
    # Get current file stats to they can be replicated after the new file is
    # moved to the destination path.
    fstat = None
    if not salt.utils.platform.is_windows():
        try:
            fstat = os.stat(dest)
        except OSError:
            pass

    # The move could fail if the dest has xattr protections, so delete the
    # temp file in this case
    try:
        shutil.move(tgt, dest)
    except Exception:  # pylint: disable=broad-except
        __clean_tmp(tgt)
        raise

    if fstat is not None:
        os.chown(dest, fstat.st_uid, fstat.st_gid)
        os.chmod(dest, fstat.st_mode)
    # If SELINUX is available run a restorecon on the file
    rcon = salt.utils.path.which("restorecon")
    if rcon:
        policy = False
        try:
            policy = salt.modules.selinux.getenforce()
        except (ImportError, CommandExecutionError):
            pass
        if policy == "Enforcing":
            with fopen(os.devnull, "w") as dev_null:
                cmd = [rcon, dest]
                subprocess.call(cmd, stdout=dev_null, stderr=dev_null)
    if os.path.isfile(tgt):
        # The temp file failed to move
        __clean_tmp(tgt)


def rename(src, dst):
    """
    On Windows, os.rename() will fail with a WindowsError exception if a file
    exists at the destination path. This function checks for this error and if
    found, it deletes the destination path first.
    """
    try:
        os.rename(src, dst)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
        try:
            os.remove(dst)
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise MinionError(
                    "Error: Unable to remove {}: {}".format(dst, exc.strerror)
                )
        os.rename(src, dst)


def process_read_exception(exc, path, ignore=None):
    """
    Common code for raising exceptions when reading a file fails

    The ignore argument can be an iterable of integer error codes (or a single
    integer error code) that should be ignored.
    """
    if ignore is not None:
        if isinstance(ignore, int):
            ignore = (ignore,)
    else:
        ignore = ()

    if exc.errno in ignore:
        return

    if exc.errno == errno.ENOENT:
        raise CommandExecutionError("{} does not exist".format(path))
    elif exc.errno == errno.EACCES:
        raise CommandExecutionError("Permission denied reading from {}".format(path))
    else:
        raise CommandExecutionError(
            "Error {} encountered reading from {}: {}".format(
                exc.errno, path, exc.strerror
            )
        )


@contextlib.contextmanager
def wait_lock(path, lock_fn=None, timeout=5, sleep=0.1, time_start=None):
    """
    Obtain a write lock. If one exists, wait for it to release first
    """
    if not isinstance(path, str):
        raise FileLockError("path must be a string")
    if lock_fn is None:
        lock_fn = path + ".w"
    if time_start is None:
        time_start = time.time()
    obtained_lock = False

    def _raise_error(msg, race=False):
        """
        Raise a FileLockError
        """
        raise FileLockError(msg, time_start=time_start)

    try:
        if os.path.exists(lock_fn) and not os.path.isfile(lock_fn):
            _raise_error("lock_fn {} exists and is not a file".format(lock_fn))

        open_flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY
        while time.time() - time_start < timeout:
            try:
                # Use os.open() to obtain filehandle so that we can force an
                # exception if the file already exists. Concept found here:
                # http://stackoverflow.com/a/10979569
                fh_ = os.open(lock_fn, open_flags)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    _raise_error(
                        "Error {} encountered obtaining file lock {}: {}".format(
                            exc.errno, lock_fn, exc.strerror
                        )
                    )
                log.trace("Lock file %s exists, sleeping %f seconds", lock_fn, sleep)
                time.sleep(sleep)
            else:
                # Write the lock file
                with os.fdopen(fh_, "w"):
                    pass
                # Lock successfully acquired
                log.trace("Write lock %s obtained", lock_fn)
                obtained_lock = True
                # Transfer control back to the code inside the with block
                yield
                # Exit the loop
                break

        else:
            _raise_error(
                "Timeout of {} seconds exceeded waiting for lock_fn {} "
                "to be released".format(timeout, lock_fn)
            )

    except FileLockError:
        raise

    except Exception as exc:  # pylint: disable=broad-except
        _raise_error(
            "Error encountered obtaining file lock {}: {}".format(lock_fn, exc)
        )

    finally:
        if obtained_lock:
            os.remove(lock_fn)
            log.trace("Write lock for %s (%s) released", path, lock_fn)


def get_umask():
    """
    Returns the current umask
    """
    ret = os.umask(0)  # pylint: disable=blacklisted-function
    os.umask(ret)  # pylint: disable=blacklisted-function
    return ret


@contextlib.contextmanager
def set_umask(mask):
    """
    Temporarily set the umask and restore once the contextmanager exits
    """
    if mask is None or salt.utils.platform.is_windows():
        # Don't attempt on Windows, or if no mask was passed
        yield
    else:
        try:
            orig_mask = os.umask(mask)  # pylint: disable=blacklisted-function
            yield
        finally:
            os.umask(orig_mask)  # pylint: disable=blacklisted-function


def fopen(*args, **kwargs):
    """
    Wrapper around open() built-in to set CLOEXEC on the fd.

    This flag specifies that the file descriptor should be closed when an exec
    function is invoked;

    When a file descriptor is allocated (as with open or dup), this bit is
    initially cleared on the new file descriptor, meaning that descriptor will
    survive into the new program after exec.

    NB! We still have small race condition between open and fcntl.
    """
    try:
        # Don't permit stdin/stdout/stderr to be opened. The boolean False
        # and True are treated by Python 3's open() as file descriptors 0
        # and 1, respectively.
        if args[0] in (0, 1, 2):
            raise TypeError("{} is not a permitted file descriptor".format(args[0]))
    except IndexError:
        pass
    binary = None
    if kwargs.pop("binary", None):
        if len(args) > 1:
            args = list(args)
            if "b" not in args[1]:
                args[1] = args[1].replace("t", "b")
                if "b" not in args[1]:
                    args[1] += "b"
        elif kwargs.get("mode"):
            if "b" not in kwargs["mode"]:
                kwargs["mode"] = kwargs["mode"].replace("t", "b")
                if "b" not in kwargs["mode"]:
                    kwargs["mode"] += "b"
        else:
            # the default is to read
            kwargs["mode"] = "rb"
    if "encoding" not in kwargs:
        # In Python 3, if text mode is used and the encoding
        # is not specified, set the encoding to 'utf-8'.
        binary = False
        if len(args) > 1:
            args = list(args)
            if "b" in args[1]:
                binary = True
        if kwargs.get("mode", None):
            if "b" in kwargs["mode"]:
                binary = True
        if not binary:
            kwargs["encoding"] = __salt_system_encoding__

    if not binary and not kwargs.get("newline", None):
        kwargs["newline"] = ""

    # Workaround callers with bad buffering setting for binary files
    if kwargs.get("buffering") == 1 and "b" in kwargs.get("mode", ""):
        log.debug(
            "Line buffering (buffering=1) isn't supported in binary mode, the default buffer size will be used"
        )
        kwargs["buffering"] = io.DEFAULT_BUFFER_SIZE

    f_handle = open(*args, **kwargs)  # pylint: disable=resource-leakage

    if is_fcntl_available():
        # modify the file descriptor on systems with fcntl
        # unix and unix-like systems only
        try:
            FD_CLOEXEC = fcntl.FD_CLOEXEC  # pylint: disable=C0103
        except AttributeError:
            FD_CLOEXEC = 1  # pylint: disable=C0103
        old_flags = fcntl.fcntl(f_handle.fileno(), fcntl.F_GETFD)
        fcntl.fcntl(f_handle.fileno(), fcntl.F_SETFD, old_flags | FD_CLOEXEC)

    return f_handle


@contextlib.contextmanager
def flopen(*args, **kwargs):
    """
    Shortcut for fopen with lock and context manager.
    """
    filename, args = args[0], args[1:]
    writing = "wa"
    with fopen(filename, *args, **kwargs) as f_handle:
        try:
            if is_fcntl_available(check_sunos=True):
                lock_type = fcntl.LOCK_SH
                if args and any([write in args[0] for write in writing]):
                    lock_type = fcntl.LOCK_EX
                fcntl.flock(f_handle.fileno(), lock_type)
            yield f_handle
        finally:
            if is_fcntl_available(check_sunos=True):
                fcntl.flock(f_handle.fileno(), fcntl.LOCK_UN)


@contextlib.contextmanager
def fpopen(*args, **kwargs):
    """
    Shortcut for fopen with extra uid, gid, and mode options.

    Supported optional Keyword Arguments:

    mode
        Explicit mode to set. Mode is anything os.chmod would accept
        as input for mode. Works only on unix/unix-like systems.

    uid
        The uid to set, if not set, or it is None or -1 no changes are
        made. Same applies if the path is already owned by this uid.
        Must be int. Works only on unix/unix-like systems.

    gid
        The gid to set, if not set, or it is None or -1 no changes are
        made. Same applies if the path is already owned by this gid.
        Must be int. Works only on unix/unix-like systems.

    """
    # Remove uid, gid and mode from kwargs if present
    uid = kwargs.pop("uid", -1)  # -1 means no change to current uid
    gid = kwargs.pop("gid", -1)  # -1 means no change to current gid
    mode = kwargs.pop("mode", None)
    with fopen(*args, **kwargs) as f_handle:
        path = args[0]
        d_stat = os.stat(path)

        if hasattr(os, "chown"):
            # if uid and gid are both -1 then go ahead with
            # no changes at all
            if (d_stat.st_uid != uid or d_stat.st_gid != gid) and [
                i for i in (uid, gid) if i != -1
            ]:
                os.chown(path, uid, gid)

        if mode is not None:
            mode_part = stat.S_IMODE(d_stat.st_mode)
            if mode_part != mode:
                os.chmod(path, (d_stat.st_mode ^ mode_part) | mode)

        yield f_handle


def safe_walk(top, topdown=True, onerror=None, followlinks=True, _seen=None):
    """
    A clone of the python os.walk function with some checks for recursive
    symlinks. Unlike os.walk this follows symlinks by default.
    """
    if _seen is None:
        _seen = set()

    # We may not have read permission for top, in which case we can't
    # get a list of the files the directory contains.  os.path.walk
    # always suppressed the exception then, rather than blow up for a
    # minor reason when (say) a thousand readable directories are still
    # left to visit.  That logic is copied here.
    try:
        # Note that listdir and error are globals in this module due
        # to earlier import-*.
        names = os.listdir(top)
    except os.error as err:
        if onerror is not None:
            onerror(err)
        return

    if followlinks:
        status = os.stat(top)
        # st_ino is always 0 on some filesystems (FAT, NTFS); ignore them
        if status.st_ino != 0:
            node = (status.st_dev, status.st_ino)
            if node in _seen:
                return
            _seen.add(node)

    dirs, nondirs = [], []
    for name in names:
        full_path = os.path.join(top, name)
        if os.path.isdir(full_path):
            dirs.append(name)
        else:
            nondirs.append(name)

    if topdown:
        yield top, dirs, nondirs
    for name in dirs:
        new_path = os.path.join(top, name)
        if followlinks or not os.path.islink(new_path):
            yield from safe_walk(new_path, topdown, onerror, followlinks, _seen)
    if not topdown:
        yield top, dirs, nondirs


def safe_rm(tgt):
    """
    Safely remove a file
    """
    try:
        os.remove(tgt)
    except OSError:
        pass


def rm_rf(path):
    """
    Platform-independent recursive delete. Includes code from
    http://stackoverflow.com/a/2656405
    """

    def _onerror(func, path, exc_info):
        """
        Error handler for `shutil.rmtree`.

        If the error is due to an access error (read only file)
        it attempts to add write permission and then retries.

        If the error is for another reason it re-raises the error.

        Usage : `shutil.rmtree(path, onerror=onerror)`
        """
        if salt.utils.platform.is_windows() and not os.access(path, os.W_OK):
            # Is the error an access error ?
            os.chmod(path, stat.S_IWUSR)
            func(path)
        else:
            raise  # pylint: disable=E0704

    if os.path.islink(path) or not os.path.isdir(path):
        os.remove(path)
    else:
        if salt.utils.platform.is_windows():
            try:
                path = salt.utils.stringutils.to_unicode(path)
            except TypeError:
                pass
        shutil.rmtree(path, onerror=_onerror)


@jinja_filter("is_empty")
def is_empty(filename):
    """
    Is a file empty?
    """
    try:
        return os.stat(filename).st_size == 0
    except OSError:
        # Non-existent file or permission denied to the parent dir
        return False


def is_fcntl_available(check_sunos=False):
    """
    Simple function to check if the ``fcntl`` module is available or not.

    If ``check_sunos`` is passed as ``True`` an additional check to see if host is
    SunOS is also made. For additional information see: http://goo.gl/159FF8
    """
    if check_sunos and salt.utils.platform.is_sunos():
        return False
    return HAS_FCNTL


def safe_filename_leaf(file_basename):
    """
    Input the basename of a file, without the directory tree, and returns a safe name to use
    i.e. only the required characters are converted by urllib.parse.quote
    If the input is a PY2 String, output a PY2 String. If input is Unicode output Unicode.
    For consistency all platforms are treated the same. Hard coded to utf8 as its ascii compatible
    windows is \\ / : * ? " < > | posix is /

    .. versionadded:: 2017.7.2

    :codeauthor: Damon Atkins <https://github.com/damon-atkins>
    """

    def _replace(re_obj):
        return urllib.parse.quote(re_obj.group(0), safe="")

    if not isinstance(file_basename, str):
        # the following string is not prefixed with u
        return re.sub(
            '[\\\\:/*?"<>|]',
            _replace,
            str(file_basename, "utf8").encode("ascii", "backslashreplace"),
        )
    # the following string is prefixed with u
    return re.sub('[\\\\:/*?"<>|]', _replace, file_basename, flags=re.UNICODE)


def safe_filepath(file_path_name, dir_sep=None):
    """
    Input the full path and filename, splits on directory separator and calls safe_filename_leaf for
    each part of the path. dir_sep allows coder to force a directory separate to a particular character

    .. versionadded:: 2017.7.2

    :codeauthor: Damon Atkins <https://github.com/damon-atkins>
    """
    if not dir_sep:
        dir_sep = os.sep
    # Normally if file_path_name or dir_sep is Unicode then the output will be Unicode
    # This code ensure the output type is the same as file_path_name
    if not isinstance(file_path_name, str) and isinstance(dir_sep, str):
        dir_sep = dir_sep.encode("ascii")  # This should not be executed under PY3
    # splitdrive only set drive on windows platform
    (drive, path) = os.path.splitdrive(file_path_name)
    path = dir_sep.join(
        [safe_filename_leaf(file_section) for file_section in path.rsplit(dir_sep)]
    )
    if drive:
        path = dir_sep.join([drive, path])
    return path


@jinja_filter("is_text_file")
def is_text(fp_, blocksize=512):
    """
    Uses heuristics to guess whether the given file is text or binary,
    by reading a single block of bytes from the file.
    If more than 30% of the chars in the block are non-text, or there
    are NUL ('\x00') bytes in the block, assume this is a binary file.
    """
    int2byte = lambda x: bytes((x,))
    text_characters = b"".join(int2byte(i) for i in range(32, 127)) + b"\n\r\t\f\b"
    try:
        block = fp_.read(blocksize)
    except AttributeError:
        # This wasn't an open filehandle, so treat it as a file path and try to
        # open the file
        try:
            with fopen(fp_, "rb") as fp2_:
                block = fp2_.read(blocksize)
        except OSError:
            # Unable to open file, bail out and return false
            return False
    if b"\x00" in block:
        # Files with null bytes are binary
        return False
    elif not block:
        # An empty file is considered a valid text file
        return True
    try:
        block.decode("utf-8")
        return True
    except UnicodeDecodeError:
        pass

    nontext = block.translate(None, text_characters)
    return float(len(nontext)) / len(block) <= 0.30


@jinja_filter("is_bin_file")
def is_binary(path):
    """
    Detects if the file is a binary, returns bool. Returns True if the file is
    a bin, False if the file is not and None if the file is not available.
    """
    if not os.path.isfile(path):
        return False
    try:
        with fopen(path, "rb") as fp_:
            try:
                data = fp_.read(2048)
                data = data.decode(__salt_system_encoding__)
                return salt.utils.stringutils.is_binary(data)
            except UnicodeDecodeError:
                return True
    except os.error:
        return False


def remove(path):
    """
    Runs os.remove(path) and suppresses the OSError if the file doesn't exist
    """
    try:
        os.remove(path)
    except OSError as exc:
        if exc.errno != errno.ENOENT:
            raise


@jinja_filter("list_files")
def list_files(directory):
    """
    Return a list of all files found under directory (and its subdirectories)
    """
    ret = set()
    ret.add(directory)
    for root, dirs, files in safe_walk(directory):
        for name in files:
            ret.add(os.path.join(root, name))
        for name in dirs:
            ret.add(os.path.join(root, name))

    return list(ret)


def st_mode_to_octal(mode):
    """
    Convert the st_mode value from a stat(2) call (as returned from os.stat())
    to an octal mode.
    """
    try:
        return oct(mode)[-4:]
    except (TypeError, IndexError):
        return ""


def normalize_mode(mode):
    """
    Return a mode value, normalized to a string and containing a leading zero
    if it does not have one.

    Allow "keep" as a valid mode (used by file state/module to preserve mode
    from the Salt fileserver in file states).
    """
    if mode is None:
        return None
    if not isinstance(mode, str):
        mode = str(mode)
    mode = mode.replace("0o", "0")
    # Strip any quotes any initial zeroes, then though zero-pad it up to 4.
    # This ensures that somethign like '00644' is normalized to '0644'
    return mode.strip('"').strip("'").lstrip("0").zfill(4)


def human_size_to_bytes(human_size):
    """
    Convert human-readable units to bytes
    """
    size_exp_map = {"K": 1, "M": 2, "G": 3, "T": 4, "P": 5}
    human_size_str = str(human_size)
    match = re.match(r"^(\d+)([KMGTP])?$", human_size_str)
    if not match:
        raise ValueError(
            "Size must be all digits, with an optional unit type (K, M, G, T, or P)"
        )
    size_num = int(match.group(1))
    unit_multiplier = 1024 ** size_exp_map.get(match.group(2), 0)
    return size_num * unit_multiplier


def backup_minion(path, bkroot):
    """
    Backup a file on the minion
    """
    dname, bname = os.path.split(path)
    if salt.utils.platform.is_windows():
        src_dir = dname.replace(":", "_")
    else:
        src_dir = dname[1:]
    if not salt.utils.platform.is_windows():
        fstat = os.stat(path)
    msecs = str(int(time.time() * 1000000))[-6:]
    if salt.utils.platform.is_windows():
        # ':' is an illegal filesystem path character on Windows
        stamp = time.strftime("%a_%b_%d_%H-%M-%S_%Y")
    else:
        stamp = time.strftime("%a_%b_%d_%H:%M:%S_%Y")
    stamp = "{}{}_{}".format(stamp[:-4], msecs, stamp[-4:])
    bkpath = os.path.join(bkroot, src_dir, "{}_{}".format(bname, stamp))
    if not os.path.isdir(os.path.dirname(bkpath)):
        os.makedirs(os.path.dirname(bkpath))
    shutil.copyfile(path, bkpath)
    if not salt.utils.platform.is_windows():
        os.chown(bkpath, fstat.st_uid, fstat.st_gid)
        os.chmod(bkpath, fstat.st_mode)


def case_insensitive_filesystem(path=None):
    """
    Detect case insensitivity on a system.

    Returns:
        bool: Flag to indicate case insensitivity

    .. versionadded:: 3004

    """
    with tempfile.NamedTemporaryFile(prefix="TmP", dir=path, delete=True) as tmp_file:
        return os.path.exists(tmp_file.name.lower())


def get_encoding(path):
    """
    Detect a file's encoding using the following:
    - Check for Byte Order Marks (BOM)
    - Check for UTF-8 Markers
    - Check System Encoding
    - Check for ascii

    Args:

        path (str): The path to the file to check

    Returns:
        str: The encoding of the file

    Raises:
        CommandExecutionError: If the encoding cannot be detected
    """

    def check_ascii(_data):
        # If all characters can be decoded to ASCII, then it's ASCII
        try:
            _data.decode("ASCII")
            log.debug("Found ASCII")
        except UnicodeDecodeError:
            return False
        else:
            return True

    def check_bom(_data):
        # Supported Python Codecs
        # https://docs.python.org/2/library/codecs.html
        # https://docs.python.org/3/library/codecs.html
        boms = [
            ("UTF-32-BE", salt.utils.stringutils.to_bytes(codecs.BOM_UTF32_BE)),
            ("UTF-32-LE", salt.utils.stringutils.to_bytes(codecs.BOM_UTF32_LE)),
            ("UTF-16-BE", salt.utils.stringutils.to_bytes(codecs.BOM_UTF16_BE)),
            ("UTF-16-LE", salt.utils.stringutils.to_bytes(codecs.BOM_UTF16_LE)),
            ("UTF-8", salt.utils.stringutils.to_bytes(codecs.BOM_UTF8)),
            ("UTF-7", salt.utils.stringutils.to_bytes("\x2b\x2f\x76\x38\x2D")),
            ("UTF-7", salt.utils.stringutils.to_bytes("\x2b\x2f\x76\x38")),
            ("UTF-7", salt.utils.stringutils.to_bytes("\x2b\x2f\x76\x39")),
            ("UTF-7", salt.utils.stringutils.to_bytes("\x2b\x2f\x76\x2b")),
            ("UTF-7", salt.utils.stringutils.to_bytes("\x2b\x2f\x76\x2f")),
        ]
        for _encoding, bom in boms:
            if _data.startswith(bom):
                log.debug("Found BOM for %s", _encoding)
                return _encoding
        return False

    def check_utf8_markers(_data):
        try:
            decoded = _data.decode("UTF-8")
        except UnicodeDecodeError:
            return False
        else:
            return True

    def check_system_encoding(_data):
        try:
            _data.decode(__salt_system_encoding__)
        except UnicodeDecodeError:
            return False
        else:
            return True

    if not os.path.isfile(path):
        raise CommandExecutionError("Not a file")
    try:
        with fopen(path, "rb") as fp_:
            data = fp_.read(2048)
    except os.error:
        raise CommandExecutionError("Failed to open file")

    # Check for Unicode BOM
    encoding = check_bom(data)
    if encoding:
        return encoding

    # Check for UTF-8 markers
    if check_utf8_markers(data):
        return "UTF-8"

    # Check system encoding
    if check_system_encoding(data):
        return __salt_system_encoding__

    # Check for ASCII first
    if check_ascii(data):
        return "ASCII"

    raise CommandExecutionError("Could not detect file encoding")



"""
Environment utilities.
"""

import os


def get_module_environment(env=None, function=None):
    """
    Get module optional environment.

    To setup an environment option for a particular module,
    add either pillar or config at the minion as follows:

    system-environment:
      modules:
        pkg:
          _:
            LC_ALL: en_GB.UTF-8
            FOO: bar
          install:
            HELLO: world
      states:
        pkg:
          _:
            LC_ALL: en_US.Latin-1
            NAME: Fred

    So this will export the environment to all the modules,
    states, returnes etc. And calling this function with the globals()
    in that context will fetch the environment for further reuse.

    Underscore '_' exports environment for all functions within the module.
    If you want to specifially export environment only for one function,
    specify it as in the example above "install".

    First will be fetched configuration, where virtual name goes first,
    then the physical name of the module overrides the virtual settings.
    Then pillar settings will override the configuration in the same order.

    :param env:
    :param function: name of a particular function
    :return: dict
    """
    result = {}
    if not env:
        env = {}
    for env_src in [env.get("__opts__", {}), env.get("__pillar__", {})]:
        fname = env.get("__file__", "")
        physical_name = os.path.basename(fname).split(".")[0]
        section = os.path.basename(os.path.dirname(fname))
        m_names = [env.get("__virtualname__")]
        if physical_name not in m_names:
            m_names.append(physical_name)
        for m_name in m_names:
            if not m_name:
                continue
            result.update(
                env_src.get("system-environment", {})
                .get(section, {})
                .get(m_name, {})
                .get("_", {})
                .copy()
            )
            if function is not None:
                result.update(
                    env_src.get("system-environment", {})
                    .get(section, {})
                    .get(m_name, {})
                    .get(function, {})
                    .copy()
                )

    return result



import logging

import pytest

import salt.ext.tornado.gen
import salt.ext.tornado.iostream
import salt.ext.tornado.tcpserver
import salt.transport.tcp
import salt.utils.msgpack

log = logging.getLogger(__name__)


@pytest.fixture
def config():
    yield {
        "master_ip": "127.0.0.1",
        "publish_port": 5679,
    }


@pytest.fixture
def server(config):
    class TestServer(salt.ext.tornado.tcpserver.TCPServer):
        send = []
        disconnect = False

        async def handle_stream(self, stream, address):
            while self.disconnect is False:
                for msg in self.send[:]:
                    msg = self.send.pop(0)
                    try:
                        await stream.write(msg)
                    except salt.ext.tornado.iostream.StreamClosedError:
                        break
                else:
                    await salt.ext.tornado.gen.sleep(1)
            stream.close()

    server = TestServer()
    try:
        yield server
    finally:
        server.disconnect = True
        server.stop()


@pytest.fixture
def client(io_loop, config):
    client = salt.transport.tcp.TCPPubClient(config.copy(), io_loop)
    try:
        yield client
    finally:
        client.close()


async def test_message_client_reconnect(io_loop, config, client, server):
    """
    Verify that the tcp MessageClient class re-sets it's unpacker after a
    stream disconnect.
    """

    server.listen(config["publish_port"])
    await client.connect(config["publish_port"])

    received = []

    def handler(msg):
        received.append(msg)

    client.on_recv(handler)

    # Prepare two packed messages
    msg = salt.utils.msgpack.dumps({"test": "test1"})
    pmsg = salt.utils.msgpack.dumps({"head": {}, "body": msg})
    assert len(pmsg) == 26
    pmsg += salt.utils.msgpack.dumps({"head": {}, "body": msg})

    # Send one full and one partial msg to the client.
    partial = pmsg[:40]
    server.send.append(partial)

    while not received:
        await salt.ext.tornado.gen.sleep(1)
    assert received == [msg]

    # The message client has unpacked one msg and there is a partial msg left in
    # the unpacker. Closing the stream now leaves the unpacker in a bad state
    # since the rest of the partil message will never be received.
    server.disconnect = True
    await salt.ext.tornado.gen.sleep(1)
    server.disconnect = False
    received = []

    # Prior to the fix for #60831, the unpacker would be left in a broken state
    # resulting in either a TypeError or BufferFull error from msgpack. The
    # rest of this test would fail.
    server.send.append(pmsg)
    while not received:
        await salt.ext.tornado.gen.sleep(1)
    assert received == [msg, msg]



"""
Module to provide Cisco UCS compatibility to Salt

:codeauthor: ``Spencer Ervin <spencer_ervin@hotmail.com>``
:maturity:   new
:depends:    none
:platform:   unix


Configuration
=============
This module accepts connection configuration details either as
parameters, or as configuration settings in pillar as a Salt proxy.
Options passed into opts will be ignored if options are passed into pillar.

.. seealso::
    :py:mod:`Cisco UCS Proxy Module <salt.proxy.cimc>`

About
=====
This execution module was designed to handle connections to a Cisco UCS server.
This module adds support to send connections directly to the device through the
rest API.

"""


import logging

import salt.proxy.cimc
import salt.utils.platform

log = logging.getLogger(__name__)

__virtualname__ = "cimc"


def __virtual__():
    """
    Will load for the cimc proxy minions.
    """
    try:
        if salt.utils.platform.is_proxy() and __opts__["proxy"]["proxytype"] == "cimc":
            return __virtualname__
    except KeyError:
        pass

    return False, "The cimc execution module can only be loaded for cimc proxy minions."


def activate_backup_image(reset=False):
    """
    Activates the firmware backup image.

    CLI Example:

    Args:
        reset(bool): Reset the CIMC device on activate.

    .. code-block:: bash

        salt '*' cimc.activate_backup_image
        salt '*' cimc.activate_backup_image reset=True

    """

    dn = "sys/rack-unit-1/mgmt/fw-boot-def/bootunit-combined"

    r = "no"

    if reset is True:
        r = "yes"

    inconfig = """<firmwareBootUnit dn='sys/rack-unit-1/mgmt/fw-boot-def/bootunit-combined'
    adminState='trigger' image='backup' resetOnActivate='{}' />""".format(
        r
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def create_user(uid=None, username=None, password=None, priv=None):
    """
    Create a CIMC user with username and password.

    Args:
        uid(int): The user ID slot to create the user account in.

        username(str): The name of the user.

        password(str): The clear text password of the user.

        priv(str): The privilege level of the user.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.create_user 11 username=admin password=foobar priv=admin

    """

    if not uid:
        raise salt.exceptions.CommandExecutionError("The user ID must be specified.")

    if not username:
        raise salt.exceptions.CommandExecutionError("The username must be specified.")

    if not password:
        raise salt.exceptions.CommandExecutionError("The password must be specified.")

    if not priv:
        raise salt.exceptions.CommandExecutionError(
            "The privilege level must be specified."
        )

    dn = "sys/user-ext/user-{}".format(uid)

    inconfig = """<aaaUser id="{0}" accountStatus="active" name="{1}" priv="{2}"
    pwd="{3}"  dn="sys/user-ext/user-{0}"/>""".format(
        uid, username, priv, password
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def get_bios_defaults():
    """
    Get the default values of BIOS tokens.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_bios_defaults

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("biosPlatformDefaults", True)

    return ret


def get_bios_settings():
    """
    Get the C240 server BIOS token values.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_bios_settings

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("biosSettings", True)

    return ret


def get_boot_order():
    """
    Retrieves the configured boot order table.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_boot_order

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("lsbootDef", True)

    return ret


def get_cpu_details():
    """
    Get the CPU product ID details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_cpu_details

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("pidCatalogCpu", True)

    return ret


def get_disks():
    """
    Get the HDD product ID details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_disks

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("pidCatalogHdd", True)

    return ret


def get_ethernet_interfaces():
    """
    Get the adapter Ethernet interface details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_ethernet_interfaces

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("adaptorHostEthIf", True)

    return ret


def get_fibre_channel_interfaces():
    """
    Get the adapter fibre channel interface details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_fibre_channel_interfaces

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("adaptorHostFcIf", True)

    return ret


def get_firmware():
    """
    Retrieves the current running firmware versions of server components.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_firmware

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("firmwareRunning", False)

    return ret


def get_hostname():
    """
    Retrieves the hostname from the device.

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_hostname

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("mgmtIf", True)

    try:
        return ret["outConfigs"]["mgmtIf"][0]["hostname"]
    except Exception as err:  # pylint: disable=broad-except
        return "Unable to retrieve hostname"


def get_ldap():
    """
    Retrieves LDAP server details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_ldap

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("aaaLdap", True)

    return ret


def get_management_interface():
    """
    Retrieve the management interface details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_management_interface

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("mgmtIf", False)

    return ret


def get_memory_token():
    """
    Get the memory RAS BIOS token.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_memory_token

    """
    ret = __proxy__["cimc.get_config_resolver_class"](
        "biosVfSelectMemoryRASConfiguration", False
    )

    return ret


def get_memory_unit():
    """
    Get the IMM/Memory unit product ID details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_memory_unit

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("pidCatalogDimm", True)

    return ret


def get_network_adapters():
    """
    Get the list of network adapters and configuration details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_network_adapters

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("networkAdapterEthIf", True)

    return ret


def get_ntp():
    """
    Retrieves the current running NTP configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_ntp

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("commNtpProvider", False)

    return ret


def get_pci_adapters():
    """
    Get the PCI adapter product ID details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_disks

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("pidCatalogPCIAdapter", True)

    return ret


def get_power_configuration():
    """
    Get the configuration of the power settings from the device. This is only available
    on some C-Series servers.

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_power_configuration

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("biosVfResumeOnACPowerLoss", True)

    return ret


def get_power_supplies():
    """
    Retrieves the power supply unit details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_power_supplies

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("equipmentPsu", False)

    return ret


def get_snmp_config():
    """
    Get the snmp configuration details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_snmp_config

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("commSnmp", False)

    return ret


def get_syslog():
    """
    Get the Syslog client-server details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_syslog

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("commSyslogClient", False)

    return ret


def get_syslog_settings():
    """
    Get the Syslog configuration settings from the system.

    .. versionadded:: 2019.2.0

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_syslog_settings

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("commSyslog", False)

    return ret


def get_system_info():
    """
    Get the system information.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_system_info

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("computeRackUnit", False)

    return ret


def get_users():
    """
    Get the CIMC users.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_users

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("aaaUser", False)

    return ret


def get_vic_adapters():
    """
    Get the VIC adapter general profile details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_vic_adapters

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("adaptorGenProfile", True)

    return ret


def get_vic_uplinks():
    """
    Get the VIC adapter uplink port details.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.get_vic_uplinks

    """
    ret = __proxy__["cimc.get_config_resolver_class"]("adaptorExtEthIf", True)

    return ret


def mount_share(
    name=None,
    remote_share=None,
    remote_file=None,
    mount_type="nfs",
    username=None,
    password=None,
):
    """
    Mounts a remote file through a remote share. Currently, this feature is supported in version 1.5 or greater.
    The remote share can be either NFS, CIFS, or WWW.

    Some of the advantages of CIMC Mounted vMedia include:
      Communication between mounted media and target stays local (inside datacenter)
      Media mounts can be scripted/automated
      No vKVM requirements for media connection
      Multiple share types supported
      Connections supported through all CIMC interfaces

      Note: CIMC Mounted vMedia is enabled through BIOS configuration.

    Args:
        name(str): The name of the volume on the CIMC device.

        remote_share(str): The file share link that will be used to mount the share. This can be NFS, CIFS, or WWW. This
        must be the directory path and not the full path to the remote file.

        remote_file(str): The name of the remote file to mount. It must reside within remote_share.

        mount_type(str): The type of share to mount. Valid options are nfs, cifs, and www.

        username(str): An optional requirement to pass credentials to the remote share. If not provided, an
        unauthenticated connection attempt will be made.

        password(str): An optional requirement to pass a password to the remote share. If not provided, an
        unauthenticated connection attempt will be made.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.mount_share name=WIN7 remote_share=10.xxx.27.xxx:/nfs remote_file=sl1huu.iso

        salt '*' cimc.mount_share name=WIN7 remote_share=10.xxx.27.xxx:/nfs remote_file=sl1huu.iso username=bob password=badpassword

    """

    if not name:
        raise salt.exceptions.CommandExecutionError("The share name must be specified.")

    if not remote_share:
        raise salt.exceptions.CommandExecutionError(
            "The remote share path must be specified."
        )

    if not remote_file:
        raise salt.exceptions.CommandExecutionError(
            "The remote file name must be specified."
        )

    if username and password:
        mount_options = " mountOptions='username={},password={}'".format(
            username, password
        )
    else:
        mount_options = ""

    dn = "sys/svc-ext/vmedia-svc/vmmap-{}".format(name)
    inconfig = """<commVMediaMap dn='sys/svc-ext/vmedia-svc/vmmap-{}' map='{}'{}
    remoteFile='{}' remoteShare='{}' status='created'
    volumeName='Win12' />""".format(
        name, mount_type, mount_options, remote_file, remote_share
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def reboot():
    """
    Power cycling the server.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.reboot

    """

    dn = "sys/rack-unit-1"

    inconfig = """<computeRackUnit adminPower="cycle-immediate" dn="sys/rack-unit-1"></computeRackUnit>"""

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def set_hostname(hostname=None):
    """
    Sets the hostname on the server.

    .. versionadded:: 2019.2.0

    Args:
        hostname(str): The new hostname to set.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.set_hostname foobar

    """
    if not hostname:
        raise salt.exceptions.CommandExecutionError("Hostname option must be provided.")

    dn = "sys/rack-unit-1/mgmt/if-1"
    inconfig = (
        """<mgmtIf dn="sys/rack-unit-1/mgmt/if-1" hostname="{}" ></mgmtIf>""".format(
            hostname
        )
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    try:
        if ret["outConfig"]["mgmtIf"][0]["status"] == "modified":
            return True
        else:
            return False
    except Exception as err:  # pylint: disable=broad-except
        return False


def set_logging_levels(remote=None, local=None):
    """
    Sets the logging levels of the CIMC devices. The logging levels must match
    the following options: emergency, alert, critical, error, warning, notice,
    informational, debug.

    .. versionadded:: 2019.2.0

    Args:
        remote(str): The logging level for SYSLOG logs.

        local(str): The logging level for the local device.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.set_logging_levels remote=error local=notice

    """

    logging_options = [
        "emergency",
        "alert",
        "critical",
        "error",
        "warning",
        "notice",
        "informational",
        "debug",
    ]

    query = ""

    if remote:
        if remote in logging_options:
            query += ' remoteSeverity="{}"'.format(remote)
        else:
            raise salt.exceptions.CommandExecutionError(
                "Remote Severity option is not valid."
            )

    if local:
        if local in logging_options:
            query += ' localSeverity="{}"'.format(local)
        else:
            raise salt.exceptions.CommandExecutionError(
                "Local Severity option is not valid."
            )

    dn = "sys/svc-ext/syslog"
    inconfig = """<commSyslog dn="sys/svc-ext/syslog"{} ></commSyslog>""".format(query)

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def set_ntp_server(server1="", server2="", server3="", server4=""):
    """
    Sets the NTP servers configuration. This will also enable the client NTP service.

    Args:
        server1(str): The first IP address or FQDN of the NTP servers.

        server2(str): The second IP address or FQDN of the NTP servers.

        server3(str): The third IP address or FQDN of the NTP servers.

        server4(str): The fourth IP address or FQDN of the NTP servers.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.set_ntp_server 10.10.10.1

        salt '*' cimc.set_ntp_server 10.10.10.1 foo.bar.com

    """

    dn = "sys/svc-ext/ntp-svc"
    inconfig = """<commNtpProvider dn="sys/svc-ext/ntp-svc" ntpEnable="yes" ntpServer1="{}" ntpServer2="{}"
    ntpServer3="{}" ntpServer4="{}"/>""".format(
        server1, server2, server3, server4
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def set_power_configuration(policy=None, delayType=None, delayValue=None):
    """
    Sets the power configuration on the device. This is only available for some
    C-Series servers.

    .. versionadded:: 2019.2.0

    Args:
        policy(str): The action to be taken when chassis power is restored after
        an unexpected power loss. This can be one of the following:

            reset: The server is allowed to boot up normally when power is
            restored. The server can restart immediately or, optionally, after a
            fixed or random delay.

            stay-off: The server remains off until it is manually restarted.

            last-state: The server restarts and the system attempts to restore
            any processes that were running before power was lost.

        delayType(str): If the selected policy is reset, the restart can be
        delayed with this option. This can be one of the following:

            fixed: The server restarts after a fixed delay.

            random: The server restarts after a random delay.

        delayValue(int): If a fixed delay is selected, once chassis power is
        restored and the Cisco IMC has finished rebooting, the system waits for
        the specified number of seconds before restarting the server. Enter an
        integer between 0 and 240.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.set_power_configuration stay-off

        salt '*' cimc.set_power_configuration reset fixed 0

    """

    query = ""
    if policy == "reset":
        query = ' vpResumeOnACPowerLoss="reset"'
        if delayType:
            if delayType == "fixed":
                query += ' delayType="fixed"'
                if delayValue:
                    query += ' delay="{}"'.format(delayValue)
            elif delayType == "random":
                query += ' delayType="random"'
            else:
                raise salt.exceptions.CommandExecutionError(
                    "Invalid delay type entered."
                )
    elif policy == "stay-off":
        query = ' vpResumeOnACPowerLoss="reset"'
    elif policy == "last-state":
        query = ' vpResumeOnACPowerLoss="last-state"'
    else:
        raise salt.exceptions.CommandExecutionError(
            "The power state must be specified."
        )

    dn = "sys/rack-unit-1/board/Resume-on-AC-power-loss"
    inconfig = """<biosVfResumeOnACPowerLoss
    dn="sys/rack-unit-1/board/Resume-on-AC-power-loss"{}>
    </biosVfResumeOnACPowerLoss>""".format(
        query
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def set_syslog_server(server=None, type="primary"):
    """
    Set the SYSLOG server on the host.

    Args:
        server(str): The hostname or IP address of the SYSLOG server.

        type(str): Specifies the type of SYSLOG server. This can either be primary (default) or secondary.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.set_syslog_server foo.bar.com

        salt '*' cimc.set_syslog_server foo.bar.com primary

        salt '*' cimc.set_syslog_server foo.bar.com secondary

    """

    if not server:
        raise salt.exceptions.CommandExecutionError(
            "The SYSLOG server must be specified."
        )

    if type == "primary":
        dn = "sys/svc-ext/syslog/client-primary"
        inconfig = """<commSyslogClient name='primary' adminState='enabled'  hostname='{}'
        dn='sys/svc-ext/syslog/client-primary'> </commSyslogClient>""".format(
            server
        )
    elif type == "secondary":
        dn = "sys/svc-ext/syslog/client-secondary"
        inconfig = """<commSyslogClient name='secondary' adminState='enabled'  hostname='{}'
        dn='sys/svc-ext/syslog/client-secondary'> </commSyslogClient>""".format(
            server
        )
    else:
        raise salt.exceptions.CommandExecutionError(
            "The SYSLOG type must be either primary or secondary."
        )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def set_user(uid=None, username=None, password=None, priv=None, status=None):
    """
    Sets a CIMC user with specified configurations.

    .. versionadded:: 2019.2.0

    Args:
        uid(int): The user ID slot to create the user account in.

        username(str): The name of the user.

        password(str): The clear text password of the user.

        priv(str): The privilege level of the user.

        status(str): The account status of the user.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.set_user 11 username=admin password=foobar priv=admin active

    """

    conf = ""
    if not uid:
        raise salt.exceptions.CommandExecutionError("The user ID must be specified.")

    if status:
        conf += ' accountStatus="{}"'.format(status)

    if username:
        conf += ' name="{}"'.format(username)

    if priv:
        conf += ' priv="{}"'.format(priv)

    if password:
        conf += ' pwd="{}"'.format(password)

    dn = "sys/user-ext/user-{}".format(uid)

    inconfig = """<aaaUser id="{0}"{1} dn="sys/user-ext/user-{0}"/>""".format(uid, conf)

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def tftp_update_bios(server=None, path=None):
    """
    Update the BIOS firmware through TFTP.

    Args:
        server(str): The IP address or hostname of the TFTP server.

        path(str): The TFTP path and filename for the BIOS image.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.tftp_update_bios foo.bar.com HP-SL2.cap

    """

    if not server:
        raise salt.exceptions.CommandExecutionError(
            "The server name must be specified."
        )

    if not path:
        raise salt.exceptions.CommandExecutionError("The TFTP path must be specified.")

    dn = "sys/rack-unit-1/bios/fw-updatable"

    inconfig = """<firmwareUpdatable adminState='trigger' dn='sys/rack-unit-1/bios/fw-updatable'
    protocol='tftp' remoteServer='{}' remotePath='{}'
    type='blade-bios' />""".format(
        server, path
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret


def tftp_update_cimc(server=None, path=None):
    """
    Update the CIMC firmware through TFTP.

    Args:
        server(str): The IP address or hostname of the TFTP server.

        path(str): The TFTP path and filename for the CIMC image.

    CLI Example:

    .. code-block:: bash

        salt '*' cimc.tftp_update_cimc foo.bar.com HP-SL2.bin

    """

    if not server:
        raise salt.exceptions.CommandExecutionError(
            "The server name must be specified."
        )

    if not path:
        raise salt.exceptions.CommandExecutionError("The TFTP path must be specified.")

    dn = "sys/rack-unit-1/mgmt/fw-updatable"

    inconfig = """<firmwareUpdatable adminState='trigger' dn='sys/rack-unit-1/mgmt/fw-updatable'
    protocol='tftp' remoteServer='{}' remotePath='{}'
    type='blade-controller' />""".format(
        server, path
    )

    ret = __proxy__["cimc.set_config_modify"](dn, inconfig, False)

    return ret



# pylint: disable=3rd-party-module-not-gated

import logging
import pathlib
import sys

from PyInstaller.utils import hooks

log = logging.getLogger(__name__)


def _filter_stdlib_tests(name):
    """
    Filter out non useful modules from the stdlib
    """
    if ".test." in name:
        return False
    if ".tests." in name:
        return False
    if ".idle_test" in name:
        return False
    return True


def _python_stdlib_path():
    """
    Return the path to the standard library folder
    """
    base_exec_prefix = pathlib.Path(sys.base_exec_prefix)
    log.info("Grabbing 'base_exec_prefix' for platform: %s", sys.platform)
    if not sys.platform.lower().startswith("win"):
        return base_exec_prefix / "lib" / "python{}.{}".format(*sys.version_info)
    return base_exec_prefix / "Lib"


def _collect_python_stdlib_hidden_imports():
    """
    Collect all of the standard library(most of it) as hidden imports.
    """
    _hidden_imports = set()

    stdlib = _python_stdlib_path()
    if not stdlib.exists():
        log.error("The path '%s' does not exist", stdlib)
        return list(_hidden_imports)

    log.info(
        "Collecting hidden imports from the python standard library at: %s",
        stdlib,
    )
    for path in stdlib.glob("*"):
        if path.is_dir():
            if path.name in (
                "__pycache__",
                "site-packages",
                "test",
                "turtledemo",
                "ensurepip",
            ):
                continue
            if path.joinpath("__init__.py").is_file():
                log.info("Collecting: %s", path.name)
                try:
                    _module_hidden_imports = hooks.collect_submodules(
                        path.name, filter=_filter_stdlib_tests
                    )
                    log.debug("Collected(%s): %s", path.name, _module_hidden_imports)
                    _hidden_imports.update(set(_module_hidden_imports))
                except Exception as exc:  # pylint: disable=broad-except
                    log.error("Failed to collect %r: %s", path.name, exc)
            continue
        if path.suffix not in (".py", ".pyc", ".pyo"):
            continue
        _hidden_imports.add(path.stem)
    log.info("Collected stdlib hidden imports: %s", sorted(_hidden_imports))
    return sorted(_hidden_imports)


def _collect_python_stdlib_dynamic_libraries():
    """
    Collect all of the standard library(most of it) dynamic libraries.
    """
    _dynamic_libs = set()

    stdlib = _python_stdlib_path()
    if not stdlib.exists():
        log.error("The path '%s' does not exist", stdlib)
        return list(_dynamic_libs)

    log.info(
        "Collecting dynamic libraries from the python standard library at: %s",
        stdlib,
    )
    for path in stdlib.glob("*"):
        if not path.is_dir():
            continue
        if path.name in (
            "__pycache__",
            "site-packages",
            "test",
            "turtledemo",
            "ensurepip",
        ):
            continue
        if path.joinpath("__init__.py").is_file():
            log.info("Collecting: %s", path.name)
            try:
                _module_dynamic_libs = hooks.collect_dynamic_libs(path.name, path.name)
                log.debug("Collected(%s): %s", path.name, _module_dynamic_libs)
                _dynamic_libs.update(set(_module_dynamic_libs))
            except Exception as exc:  # pylint: disable=broad-except
                log.error("Failed to collect %r: %s", path.name, exc)
    log.info("Collected stdlib dynamic libs: %s", sorted(_dynamic_libs))
    return sorted(_dynamic_libs)


def _filter_submodules(name):
    # this should never happen, but serves as a place-holder for when/if we have to filter
    if not name.startswith("salt"):
        return False
    return True


# Collect Salt datas, binaries(should be None) and hidden imports
SALT_DATAS, SALT_BINARIES, SALT_HIDDENIMPORTS = hooks.collect_all(
    "salt",
    include_py_files=True,
    filter_submodules=_filter_submodules,
)

# In case there's salt-extensions installed, collect their datas and hidden imports
SALT_EXTENSIONS_DATAS, SALT_EXTENSIONS_HIDDENIMPORTS = hooks.collect_entry_point(
    "salt.loader"
)


# PyInstaller attributes
datas = sorted(set(SALT_DATAS + SALT_EXTENSIONS_DATAS))
binaries = sorted(set(SALT_BINARIES))
hiddenimports = sorted(
    set(
        SALT_HIDDENIMPORTS
        + SALT_EXTENSIONS_HIDDENIMPORTS
        + _collect_python_stdlib_hidden_imports()
    )
)



import pytest

import salt.netapi
from salt.exceptions import EauthAuthenticationError, SaltInvocationError
from tests.support.mock import patch


@pytest.fixture
def client(salt_minion, salt_sub_minion, client_config):
    return salt.netapi.NetapiClient(client_config)


@pytest.mark.slow_test
def test_local(client, auth_creds, salt_minion, salt_sub_minion):
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **auth_creds}

    with patch.dict(client.opts, {"netapi_enable_clients": ["local"]}):
        ret = client.run(low)
    assert ret == {salt_minion.id: True, salt_sub_minion.id: True}


@pytest.mark.slow_test
def test_local_batch(client, auth_creds, salt_minion, salt_sub_minion):
    low = {"client": "local_batch", "tgt": "*", "fun": "test.ping", **auth_creds}

    with patch.dict(client.opts, {"netapi_enable_clients": ["local_batch"]}):
        ret = client.run(low)
    assert ret
    # local_batch returns a generator
    ret = list(ret)
    assert ret
    assert {salt_minion.id: True} in ret
    assert {salt_sub_minion.id: True} in ret


def test_local_async(client, auth_creds, salt_minion, salt_sub_minion):
    low = {"client": "local_async", "tgt": "*", "fun": "test.ping", **auth_creds}

    with patch.dict(client.opts, {"netapi_enable_clients": ["local_async"]}):
        ret = client.run(low)

    assert "jid" in ret
    assert sorted(ret["minions"]) == sorted([salt_minion.id, salt_sub_minion.id])


def test_local_unauthenticated(client):
    low = {"client": "local", "tgt": "*", "fun": "test.ping"}

    with patch.dict(client.opts, {"netapi_enable_clients": ["local"]}):
        with pytest.raises(EauthAuthenticationError):
            client.run(low)


def test_local_disabled(client, auth_creds):
    low = {"client": "local", "tgt": "*", "fun": "test.ping", **auth_creds}

    ret = None
    with pytest.raises(SaltInvocationError):
        ret = client.run(low)

    assert ret is None


def test_local_batch_disabled(client, auth_creds):
    low = {"client": "local_batch", "tgt": "*", "fun": "test.ping", **auth_creds}

    ret = None
    with pytest.raises(SaltInvocationError):
        ret = client.run(low)

    assert ret is None


def test_local_subset_disabled(client, auth_creds):
    low = {
        "client": "local_subset",
        "tgt": "*",
        "fun": "test.ping",
        "subset": 1,
        **auth_creds,
    }

    ret = None
    with pytest.raises(SaltInvocationError):
        ret = client.run(low)

    assert ret is None


@pytest.mark.slow_test
def test_wheel(client, auth_creds):
    low = {"client": "wheel", "fun": "key.list_all", **auth_creds}

    with patch.dict(client.opts, {"netapi_enable_clients": ["wheel"]}):
        ret = client.run(low)

    assert "tag" in ret
    assert "data" in ret
    assert "jid" in ret["data"]
    assert "tag" in ret["data"]
    assert "return" in ret["data"]
    assert "local" in ret["data"]["return"]
    assert {"master.pem", "master.pub"}.issubset(set(ret["data"]["return"]["local"]))


@pytest.mark.slow_test
def test_wheel_async(client, auth_creds):
    low = {"client": "wheel_async", "fun": "key.list_all", **auth_creds}

    with patch.dict(client.opts, {"netapi_enable_clients": ["wheel_async"]}):
        ret = client.run(low)
    assert "jid" in ret
    assert "tag" in ret


def test_wheel_unauthenticated(client):
    low = {"client": "wheel", "tgt": "*", "fun": "test.ping"}

    with patch.dict(client.opts, {"netapi_enable_clients": ["wheel"]}):
        with pytest.raises(EauthAuthenticationError):
            client.run(low)


def test_wheel_disabled(client, auth_creds):
    low = {"client": "wheel", "tgt": "*", "fun": "test.ping", **auth_creds}

    ret = None
    with pytest.raises(SaltInvocationError):
        ret = client.run(low)

    assert ret is None


def test_wheel_async_disabled(client, auth_creds):
    low = {"client": "wheel_async", "tgt": "*", "fun": "test.ping", **auth_creds}

    ret = None
    with pytest.raises(SaltInvocationError):
        ret = client.run(low)

    assert ret is None


def test_runner_unauthenticated(client):
    low = {"client": "runner", "tgt": "*", "fun": "test.ping"}

    with patch.dict(client.opts, {"netapi_enable_clients": ["runner"]}):
        with pytest.raises(EauthAuthenticationError):
            client.run(low)


def test_runner_disabled(client, auth_creds):
    low = {"client": "runner", "tgt": "*", "fun": "test.ping", **auth_creds}

    ret = None
    with pytest.raises(SaltInvocationError):
        ret = client.run(low)

    assert ret is None



"""
Test the ssh module
"""

import logging
import pathlib
import re
import shutil
import subprocess
import textwrap

import pytest

from tests.support.runtests import RUNTIME_VARS

GPG_TEST_PRIV_KEY = """-----BEGIN PGP PRIVATE KEY BLOCK-----
Version: GnuPG v2.0.22 (GNU/Linux)
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=mNsA
-----END PGP PRIVATE KEY BLOCK-----
"""

GPG_TEST_PUB_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v2.0.22 (GNU/Linux)

mQENBF7qopoBCACVyX0rjtcbh1CMu0Od/luO14VvgEZ+u+MJaioiEN6NMU0SscBn
A4/+BSu/f6DPc9MgJBM6AI0O4a7OM00kwZHODT3TykLEJKbL+taObkHcS9j5b5iA
ife+56W8KLEZpYk6uhk9LXMKilc7R01qXgLrDgqx9kGIe/5bC/AtCEW656zOWRPn
1p55oWW2toM7pVkdsOG1UZ43uZq1+iTSuxe+FcH0NU0yPBISFa+x5kuKP+xBa8Ou
n206+NGtnQQJjlbPtce/QrJU3mQP8xgBZBdU7NevxYTyQrkJXwUPqCXevqCU4KKX
c+I+nAJsrvnbdcnBWe5bUvVnpKMOLguxeKtzABEBAAG0YVNhbHRTdGFjayBQYWNr
YWdpbmcgVGVhbSAoVGhpcyBrZXkgaXMgZm9yIHRlc3RpbmcgcGFja2FnZSBzaWdu
aW5nIG9ubHkpIDxwYWNrYWdpbmdAc2FsdHN0YWNrLmNvbT6JATkEEwECACMFAl7q
opoCGwMHCwkIBwMCAQYVCAIJCgsEFgIDAQIeAQIXgAAKCRAfceQnFUySQdC1B/wL
09fhDkhKzgs6jpcBSKHSxMR/dD4l0ehmw82T4mqpoYRGOsWn+b7fCJTGQIWXL/Rr
+hPqV1tKBVRsk3WiRL35xdFThPsWLjjNFz06cZ9VGhp07udimcpcUAQZHMT+OGnC
hDUNYQkTp17u529KvN4Nb3byo8GtjtYSl1V9tPbU8RpMO+c/Kx1Ope1KT8EO+/QE
H4jIwFRh/h9PSbblB18gXCZwvQB5lKiXjWlhG7/FCQpzPEKs/SDgopURD3rgjxie
cpfcLxvGQ4xSdukTLzEXzqM1c6lEd3GL9FsNIfZsm15AujiSzsx91hGsjN6TplCY
EF4Vly94p6ChBtaqpxyBuQENBF7qopoBCADhL6siQvvm18xxQeUC5CKN2A3pfSDi
iehf8/etfSCYls4h1KlZPvcHQz8h45yKvtQF29Ly1HZiFZp+Un+WLx4p+bVp/Ysi
A+L8+McvZhCmXuL5/jui1YX/Lv2gd9nUNTm0UG9I2grT6MsYkK/t0nv/VtocMPkS
6Kc0cuaZnKKKDG4GIhPWbEmkba+YJN/7R6bXaNfvg5BQlvYGI9nnX/OTzqBbYWDG
49nIb5mTGz7tZNXd/l6muana31UVRZEGybl0VJ1NCVEE6uAdEQAog/HiZbuI1ona
6y1lbpGOTAoPmj5bQnx/DAAjfDOl0oRrTOlucStAF8yyv4MPzbrzZd6nABEBAAGJ
AR8EGAECAAkFAl7qopoCGwwACgkQH3HkJxVMkkEl0QgAgqVDZywU6it+X29dZd0y
EwjxtDzu4W2kD86484sqkwtytpAGG6f/UF4s90oIvQx579JdwBPDjckGCLT+SE5i
o0QMQZIRFCSTI2WSil8nnR/HEnT88lNgeMme8MvfP0vjijZpzHaGPOv3dLuHNB0S
QCvz8Skwe6yGgKhUnKX6p2WUG3MTJpA1Ccd/J1JLojtYnkL0U7576icMfZuBlZlc
CBDJpTNeaNG/OGaHwukI1kHM8ZW+hh+8YlFF6rlv04DLrbDHEKQZX2WKT4rNAOFp
mCCvdH7g1yJm9b2crlJuJ3VZl5KI9hQdGpBCcLhcMn3736n0fOURlmx8J6JF/kAu
yQ==
=6abY
-----END PGP PUBLIC KEY BLOCK-----
"""

GPG_TEST_KEY_PASSPHRASE = "saltstacktestingpackages"
GPG_TEST_KEY_ID = "154C9241"

GPG_AGENT_CONF = """default-cache-ttl 600
default-cache-ttl-ssh 600
max-cache-ttl 600
max-cache-ttl-ssh 600
allow-preset-passphrase
daemon
debug-all
## debug-pinentry
log-file /root/gpg-agent.log
verbose
# PIN entry program
#pinentry-program /usr/bin/pinentry-gtk
#pinentry-timeout 30
#allow-loopback-pinentry
"""
REPO_NAMED_RPM = "hello-2.10-1.el7.x86_64.rpm"

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        True, reason="This test never ran and it should be revisited by David Murphy"
    ),
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_if_binaries_missing(
        "gpg",
        "rpm",
        "rpmbuild",
        "rpmsign",
        "createrepo",
        "gpgconf",
        "gpg-agent",
        "/usr/libexec/gpg-preset-passphrase",
    ),
    pytest.mark.requires_salt_modules("gpg", "pkgbuild.make_repo"),
]


@pytest.fixture(scope="module")
def pillar_tree(base_env_pillar_tree_root_dir, salt_minion, salt_call_cli):
    top_file = """
    base:
      '{}':
        - packaging
    """.format(
        salt_minion.id
    )
    packaging_pillar_file = """
    gpg_pkg_priv_key: |
      {}

    gpg_pkg_priv_keyname: gpg_pkg_key.pem


    gpg_pkg_pub_key: |
      {}

    gpg_pkg_pub_keyname: gpg_pkg_key.pub

    gpg_passphrase: "{}"
    """.format(
        textwrap.indent(GPG_TEST_PRIV_KEY, "      ").lstrip(),
        textwrap.indent(GPG_TEST_PUB_KEY, "      ").lstrip(),
        GPG_TEST_KEY_PASSPHRASE,
    )
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    packaging_tempfile = pytest.helpers.temp_file(
        "packaging.sls", packaging_pillar_file, base_env_pillar_tree_root_dir
    )

    try:
        with top_tempfile, packaging_tempfile:
            ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
            assert ret.returncode == 0
            assert ret.data is True
            yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


def _testrpm_signed(abs_path_named_rpm):
    try:
        rpm_chk_sign = subprocess.Popen(
            ["rpm", "--checksig", "-v", abs_path_named_rpm],
            shell=False,
            close_fds=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).communicate()[0]
    except OSError:
        return False
    if not rpm_chk_sign:
        log.debug("problem checking signatures on rpm")
        return False

    test_string = GPG_TEST_KEY_ID.lower() + ": OK"
    CHECK_KEYID_OK = re.compile(test_string, re.M)
    retrc = CHECK_KEYID_OK.search(rpm_chk_sign.decode())
    log.debug(
        "signed checking, found test_string '%s' in rpm_chk_sign '%s', return code '%s'",
        test_string,
        rpm_chk_sign,
        retrc,
    )
    if retrc:
        return True
    return False


@pytest.fixture
def gpghome(tmp_path):
    _gpghome = tmp_path / ".gnupg"
    _gpghome.mkdir()
    gpgconf = _gpghome / "gpg.conf"
    gpgconf.write_text("use-agent\n")
    agentconf = _gpghome / "gpg-agent.conf"
    agentconf.write_text(GPG_AGENT_CONF)
    gpg_pkg_key_pem_path = _gpghome / "gpg_pkg_key.pem"
    gpg_pkg_key_pem_path.write_text(GPG_TEST_PRIV_KEY)
    gpg_pkg_key_pem_path.chmod(0o600)
    gpg_pkg_key_pub_path = _gpghome / "gpg_pkg_key.pub"
    gpg_pkg_key_pub_path.write_text(GPG_TEST_PUB_KEY)
    gpg_pkg_key_pub_path.chmod(0o600)
    _gpghome.chmod(0o600)
    return _gpghome


@pytest.fixture
def repodir(tmp_path):
    _repodir = tmp_path / "repo"
    _repodir.mkdir()
    shutil.copy(
        str(pathlib.Path(RUNTIME_VARS.FILES) / REPO_NAMED_RPM),
        str(_repodir / REPO_NAMED_RPM),
    )
    return _repodir


def gpg_agent_ids(value):
    if value:
        return "with-gpg-agent"
    return "without-gpg-agent"


@pytest.fixture(params=(True, False), ids=gpg_agent_ids)
def gpg_agent(request, gpghome):

    gpg_version_proc = subprocess.run(  # nosec
        "gpgconf --version | head -n 1 | awk '{ print $3 }'",
        shell=True,
        stdout=subprocess.PIPE,
        check=True,
        universal_newlines=True,
    )
    if tuple(int(p) for p in gpg_version_proc.stdout.split(".")) >= (2, 1):
        kill_option_supported = True
    else:
        kill_option_supported = False

    # Kill gpg-agent
    if kill_option_supported:
        subprocess.run(["gpgconf", "--kill", "gpg-agent"], check=True)
    else:
        try:
            pidof_gpg_agent = subprocess.run(
                ["pidof", "gpg-agent"],
                check=True,
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )
        except subprocess.CalledProcessError as exc:
            # Not running
            pass
        else:
            subprocess.run(
                ["kill", "-KILL", pidof_gpg_agent.stdout.strip()], check=True
            )

    if request.param is False:
        # No gpg agent
        yield
        return

    try:
        # launch gpg-agent
        gpg_tty_info_path = gpghome / "gpg_tty_info"
        gpg_agent_cmd = (
            "gpg-agent --homedir {} --allow-preset-passphrase --max-cache-ttl 600"
            " --daemon".format(gpghome)
        )
        echo_gpg_tty_cmd = (
            "GPG_TTY=$(tty) ; export GPG_TTY ; echo $GPG_TTY=$(tty) > {}".format(
                gpg_tty_info_path
            )
        )
        subprocess.run(  # nosec
            "{}; {}".format(gpg_agent_cmd, echo_gpg_tty_cmd), shell=True, check=True
        )
        yield
    finally:
        # Kill gpg-agent
        if kill_option_supported:
            subprocess.run(["gpgconf", "--kill", "gpg-agent"], check=True)
        else:
            try:
                pidof_gpg_agent = subprocess.run(
                    ["pidof", "gpg-agent"],
                    check=True,
                    stdout=subprocess.PIPE,
                    universal_newlines=True,
                )
            except subprocess.CalledProcessError as exc:
                # Not running
                pass
            else:
                subprocess.run(
                    ["kill", "-KILL", pidof_gpg_agent.stdout.strip()], check=True
                )


@pytest.mark.slow_test
def test_make_repo(grains, gpghome, repodir, gpg_agent, salt_call_cli, pillar_tree):
    """
    test make repo, signing rpm
    """
    if not (grains["os_family"] == "RedHat" or grains["os_family"] == "Amazon"):
        pytest.skip("TODO: test not configured for {}".format(grains["os_family"]))

    if grains["os_family"] == "RedHat" and grains["osmajorrelease"] >= 8:
        pytest.skip(
            "TODO: test not configured for {} and major release {}".format(
                grains["os_family"], grains["osmajorrelease"]
            )
        )

    if grains["os_family"] == "Amazon" and grains["osmajorrelease"] != 2:
        pytest.skip(
            "TODO: test not configured for {} and major release {}".format(
                grains["os_family"], grains["osmajorrelease"]
            )
        )

    ret = salt_call_cli.run("pillar.data")
    assert ret.returncode == 0
    assert ret.data
    pillar = ret.data
    assert pillar["gpg_passphrase"] == GPG_TEST_KEY_PASSPHRASE
    assert pillar["gpg_pkg_pub_keyname"] == "gpg_pkg_key.pub"

    # sign rpm and create repo
    ret = salt_call_cli.run(
        "pkgbuild.make_repo",
        repodir=str(repodir),
        keyid=GPG_TEST_KEY_ID,
        env=None,
        use_passphrase=True,
        gnupghome=str(gpghome),
        runas="root",
        timeout=15.0,
    )
    assert ret.returncode == 0
    assert ret.data

    test_rpm_path = repodir / REPO_NAMED_RPM
    assert _testrpm_signed(test_rpm_path)
    test_repodata_path = repodir / "repodata"
    assert test_repodata_path.is_dir()
    test_repomd_xml_path = test_repodata_path / "repomd.xml"
    assert test_repomd_xml_path.is_file()



"""
Manage Rackspace Queues
=======================

.. versionadded:: 2015.5.0

Create and destroy Rackspace queues. Be aware that this interacts with
Rackspace's services, and so may incur charges.

This module uses ``pyrax``, which can be installed via package, or pip.
This module is greatly inspired by boto_* modules from SaltStack code source.

.. code-block:: yaml

    myqueue:
        pyrax_queues.present:
            - provider: my-pyrax

    myqueue:
        pyrax_queues.absent:
            - provider: my-pyrax
"""

import salt.utils.openstack.pyrax as suop


def __virtual__():
    """
    Only load if pyrax is available.
    """
    return suop.HAS_PYRAX


def present(name, provider):
    """
    Ensure the RackSpace queue exists.

    name
        Name of the Rackspace queue.

    provider
        Salt Cloud Provider
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_present = next(
        iter(
            __salt__["cloud.action"]("queues_exists", provider=provider, name=name)[
                provider
            ].values()
        )
    )

    if not is_present:
        if __opts__["test"]:
            msg = "Rackspace queue {} is set to be created.".format(name)
            ret["comment"] = msg
            ret["result"] = None
            return ret
        created = __salt__["cloud.action"](
            "queues_create", provider=provider, name=name
        )
        if created:
            queue = __salt__["cloud.action"](
                "queues_show", provider=provider, name=name
            )
            ret["changes"]["old"] = {}
            ret["changes"]["new"] = {"queue": queue}
        else:
            ret["result"] = False
            ret["comment"] = "Failed to create {} Rackspace queue.".format(name)
            return ret
    else:
        ret["comment"] = "{} present.".format(name)

    return ret


def absent(name, provider):
    """
    Ensure the named Rackspace queue is deleted.

    name
        Name of the Rackspace queue.

    provider
        Salt Cloud provider
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    is_present = next(
        iter(
            __salt__["cloud.action"]("queues_exists", provider=provider, name=name)[
                provider
            ].values()
        )
    )

    if is_present:
        if __opts__["test"]:
            ret["comment"] = "Rackspace queue {} is set to be removed.".format(name)
            ret["result"] = None
            return ret
        queue = __salt__["cloud.action"]("queues_show", provider=provider, name=name)
        deleted = __salt__["cloud.action"](
            "queues_delete", provider=provider, name=name
        )
        if deleted:
            ret["changes"]["old"] = queue
            ret["changes"]["new"] = {}
        else:
            ret["result"] = False
            ret["comment"] = "Failed to delete {} Rackspace queue.".format(name)
    else:
        ret["comment"] = "{} does not exist.".format(name)

    return ret
